/*
Model we want to test:
        We expect the distance from the Z axis to the cohesin
atoms in the model to settle down to 220 nanometers (which is
where we placed them in the original model).  This code will
compute the min, max, mean, and variance of the distance from
the Z axis for all points that are part of the cohesin (material
ID #4) in a given animation file.
*/

/* XXX
 * Problem: If the last time step is truncated and doesn't spit
 * out all of the points, then we get a bogus last line.
 * Quick fix would be to check that we got some cohesin points.  Better
 * fix would be to check that we got the same number of points.
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <ctype.h>

typedef std::vector<double> POINT;
typedef std::vector<POINT>  POINTS_FOR_A_GIVEN_TIME;

// Find the mean distance from the Z axis for all of
// the points at a given time.
static double find_mean_dist_from_Z(const POINTS_FOR_A_GIVEN_TIME &pts)
{
    // If we have no points, return zero.
    double mean = 0;
    if (pts.size() == 0) {
        return mean;
    }

    // Compute the mean distance of the all the points
    size_t i;
    for (i = 0; i < pts.size(); i++) {
        mean += sqrt(pts[i][0]*pts[i][0] + pts[i][1]*pts[i][1]);
    }
    mean /= pts.size();

    return mean;
}

// Find the minimum distance from the Z axis for all of
// the points at a given time.
static double find_min_dist_from_Z(const POINTS_FOR_A_GIVEN_TIME &pts)
{
    // If we have no points, return 1e6.
    double min = 1e6;
    if (pts.size() == 0) {
        return min;
    }

    // Compute the min distance of the all the points
    size_t i;
    for (i = 0; i < pts.size(); i++) {
        double dist = sqrt(pts[i][0]*pts[i][0] + pts[i][1]*pts[i][1]);
        if (dist < min) { min = dist; }
    }

    return min;
}

// Find the maximum distance from the Z axis for all of
// the points at a given time.
static double find_max_dist_from_Z(const POINTS_FOR_A_GIVEN_TIME &pts)
{
    // If we have no points, return 0.
    double max = 0;
    if (pts.size() == 0) {
        return max;
    }

    // Compute the min distance of the all the points
    size_t i;
    for (i = 0; i < pts.size(); i++) {
        double dist = sqrt(pts[i][0]*pts[i][0] + pts[i][1]*pts[i][1]);
        if (dist > max) { max = dist; }
    }

    return max;
}

void Usage(const char *name)
{
    fprintf(stderr,"Usage: %s [-trim DIST] ColorFileName [InputFileName]\n", name);
    fprintf(stderr,"       -trim: Trim points less than DIST from the center\n");
    fprintf(stderr,"       ColorFileName: File containing the material identifiers for each point\n");
    fprintf(stderr,"       InputFileName: File containing the time steps (default stdin)\n");
    exit(-1);
}

int main(int argc, const char *argv[])
{
    // "Color" file containing the material indices for each point
    FILE *colorfile = NULL;
    const char *colorfile_name = "";

    // Get the input and output files (by default, just standard in and out)
    FILE *infile = stdin;
    const char *infile_name = "";
    FILE *outfile = stdout;

    // If points are closer than this to the center, trim them from the
    // statistics.  This is because some models had colored the fixed
    // end-points of the DNA the same color as the cohesin rings.
    double trim_dist = 0.0;

    // Parse the command line.
    size_t i;
    unsigned real_params = 0;
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i], "-trim")) {
            if (++i >= argc) { Usage(argv[0]); }
            trim_dist = atof(argv[i]);
        } else if (argv[i][0] == '-') {
            Usage(argv[0]);
        } else switch (real_params++) {
            case 0:
                colorfile_name = argv[i];
                break;
            case 1:
                infile_name = argv[i];
                break;
            default:
                Usage(argv[0]);
        }
    }

    // Open the input files.
    if (strlen(colorfile_name) == 0) {
        Usage(argv[0]);
    } else if ( NULL == (colorfile = fopen(colorfile_name, "r"))) {
        perror("Cannot open color file for reading");
        exit(-2);
    }
    if ( (strlen(infile_name) > 0) && (NULL == (infile = fopen(infile_name, "r"))) ) {
        perror("Cannot open input file for reading");
        exit(-2);
    }

    // Read the color file.  Each line consists of a single integer that gives
    // the material index for a specific point.
    std::vector<int>   colors;
    char    line[4096];
    line[sizeof(line)-1] = '\0';    // Ensure null-termination
    int color;
    while (fscanf(colorfile, "%d", &color) == 1) {
        colors.push_back(color);
    }

    // We repeat reading in blocks of points, each of which is preceded
    // by a line that says "Time T", where T is a floating-point number.
    // Then there are a number of lines of X Y Z which are floating-point
    // coordinates. Then there is a blank line after each block of points.
    // We only add points whose corresponding color index equals 4 (the
    // one for cohesin).
    std::vector<POINTS_FOR_A_GIVEN_TIME>    times;

	long int line_num = 0;
	bool foundCoordinates = false;
    while (true) {
		line_num++;
        // Read and parse the "Time" line.
        if (fgets(line, sizeof(line)-1, infile) == NULL) {
            // Done reading lines, we got them all.
            break;
        }
        double t;
		if (sscanf(line, "Time %le", &t) <= 0) {
			if (foundCoordinates) {
				// If the line is empty, ignore it and move on.
				// Otherwise,we have a bad time line
				char c;
				int i = 0;
				while (line[i])
				{
					c = line[i];
					if (!isspace(c)) {
						fprintf(stderr, "Bad Time line %ld: %s\n", line_num, line);
						return(-1);
					}
					i++;
				}
			}
			else {
				// haven't arrived at the first timestep of coordinates
				continue;
			}
        }
		else if (!foundCoordinates) {
			foundCoordinates = true;
		}

        // Read and parse point lines, adding them to our time.
        // Only add points that correspond to cohesin.
        POINTS_FOR_A_GIVEN_TIME points;
        times.push_back(points);
        size_t point_count = 0;
        while (true) {
            // Read and parse the next line, which may be a points line.
			line_num++;
            if (fgets(line, sizeof(line)-1, infile) == NULL) {
                // Done reading lines, we got them all.
                break;
            }
            double x,y,z;
            if (sscanf(line, "%le %le %le", &x, &y, &z) != 3) {
                // No more points for this line
                break;
            }
\
            // Trim points closer than the trim distance from the
            // Z axis
            if (trim_dist > sqrt(x*x + y*y)) {
                point_count++;
                continue;
            }

			// Find out the material color for this point. If we don't
			// have one, all the "important" points have been found.
			// Previously all points were included in the color file, even past
			// the last important point, so this would be considered an error.
			// Can switch back if it causes problems to use color files without
			// all the points.
			if (point_count >= colors.size()) {
				continue;
				//fprintf(stderr, "More points in time %d than in the color file\n");
				// exit(-3);
			}
            else if (colors[point_count] == 4) {
                POINT p;
                p.push_back(x);
                p.push_back(y);
                p.push_back(z);
                times.back().push_back(p);
            }
            point_count++;
        }
    }
    fclose(infile);

    // Compute statistics on how far the points are away from the Z axis.
    // Print out the radius of gyration at each time step.
    fprintf(outfile, "Mean dist, Min dist, Max dist\n");
    for (i = 0; i < times.size(); i++) {
        fprintf(outfile, "%le, %le, %le\n",
                find_mean_dist_from_Z(times[i]),
                find_min_dist_from_Z(times[i]),
                find_max_dist_from_Z(times[i]) );
    }

    fclose(outfile);
    return 0;
}
